import pytest

import orderly_web
import orderly_web.cli
from orderly_web.cli import string_to_dict
from orderly_web.status import print_status


def test_cli_parse_start():
    target, args = orderly_web.cli.parse_args(["start", "path"])
    assert target == orderly_web.start
    assert args == ("path", None, [], False)


def test_cli_parse_start_with_extra():
    target, args = orderly_web.cli.parse_args(
        ["start", "path", "--extra=extra.yml"])
    assert target == orderly_web.start
    assert args == ("path", "extra.yml", [], False)


def test_cli_parse_start_with_options():
    target, args = orderly_web.cli.parse_args(
        ["start", "path", "--option=a=x", "--option=b.c=y"])
    assert target == orderly_web.start
    options = [{'a': 'x'}, {'b': {'c': 'y'}}]
    assert args == ("path", None, options, False)


def test_cli_parse_status():
    target, args = orderly_web.cli.parse_args(["status", "path"])
    assert target == print_status
    assert args == ("path",)


def test_cli_parse_stop():
    target, args = orderly_web.cli.parse_args(["stop", "path"])
    assert target == orderly_web.stop
    assert args == ("path", False, False, False, False, None, [])

    target, args = orderly_web.cli.parse_args(["stop", "path", "--kill"])
    assert target == orderly_web.stop
    assert args == ("path", True, False, False, False, None, [])

    target, args = orderly_web.cli.parse_args(["stop", "path", "--network"])
    assert args == ("path", False, True, False, False, None, [])

    target, args = orderly_web.cli.parse_args(["stop", "path", "--volumes"])
    assert args == ("path", False, False, True, False, None, [])

    target, args = orderly_web.cli.parse_args(["stop", "path", "--force",
                                               "--extra=./extra_path",
                                               "--option=k=v"])
    assert args == ("path", False, False, False, True, "./extra_path",
                    [{"k": "v"}])


def test_cli_parse_add_users():
    args = ["admin",
            "path",
            "add-users",
            "test.user@example.com",
            "another.user@email.com"]
    target, args = orderly_web.cli.parse_args(args)
    assert target == orderly_web.add_users
    assert args == ("path", ["test.user@example.com",
                             "another.user@email.com"])


def test_cli_parse_add_groups():
    args = ["admin",
            "path",
            "add-groups",
            "admin",
            "funders"]
    target, args = orderly_web.cli.parse_args(args)
    assert target == orderly_web.add_groups
    assert args == ("path", ["admin", "funders"])


def test_cli_parse_add_members():
    args = ["admin",
            "path",
            "add-members",
            "admin",
            "test.user@email.com",
            "another.user@email.com"]
    target, args = orderly_web.cli.parse_args(args)
    assert target == orderly_web.add_members
    assert args == ("path", "admin", ["test.user@email.com",
                                      "another.user@email.com"])


def test_cli_parse_grant():
    args = ["admin",
            "path",
            "grant",
            "admin",
            "*/reports.read",
            "*/reports.run"]
    target, args = orderly_web.cli.parse_args(args)
    assert target == orderly_web.grant
    assert args == ("path", "admin", ["*/reports.read", "*/reports.run"])


def test_cli_parse_validate_options():
    msg = "Invalid option 'foo', expected option in form key=value"
    with pytest.raises(Exception, match=msg):
        orderly_web.cli.parse_args(["start", "path", "--option=foo"])
    msg = "Invalid option 'a=b=c', expected option in form key=value"
    with pytest.raises(Exception, match=msg):
        orderly_web.cli.parse_args(["start", "path", "--option=a=b=c"])


def test_string_to_dict():
    assert string_to_dict("a=x") == {"a": "x"}
    assert string_to_dict("a.b=x") == {"a": {"b": "x"}}
    assert string_to_dict("a.b.c=x") == {"a": {"b": {"c": "x"}}}
    assert string_to_dict("a.b.c=True") == {"a": {"b": {"c": True}}}
    assert string_to_dict("a.b.c=1") == {"a": {"b": {"c": 1}}}
    assert string_to_dict("a.b.c=1.23") == {"a": {"b": {"c": 1.23}}}
    msg = "Invalid option 'foo', expected option in form key=value"
    with pytest.raises(Exception, match=msg):
        string_to_dict("foo")
    msg = "Invalid option 'a=b=c', expected option in form key=value"
    with pytest.raises(Exception, match=msg):
        string_to_dict("a=b=c")
    msg = "Invalid value '{}' - expected simple type"
    with pytest.raises(Exception, match=msg):
        string_to_dict("a={}")



"""Usage:
  orderly-web start <path> [--extra=PATH] [--option=OPTION]... [--pull]
  orderly-web status <path>
  orderly-web stop <path> [--volumes] [--network] [--kill] [--force]
    [--extra=PATH] [--option=OPTION]...
  orderly-web admin <path> add-users <email>...
  orderly-web admin <path> add-groups <name>...
  orderly-web admin <path> add-members <group> <email>...
  orderly-web admin <path> grant <group> <permission>...

Options:
  --extra=PATH     Path, relative to <path>, of yml file of additional
                   configuration
  --option=OPTION  Additional configuration options, in the form key=value
                   Use dots in key for hierarchical structure, e.g., a.b=value
                   This argument may be repeated to provide multiple arguments
  --pull           Pull images before starting
  --volumes        Remove volumes (WARNING: irreversible data loss)
  --network        Remove network
  --kill           Kill the containers (faster, but possible db corruption)
  --force          Force stop even if containers are corrupted and cannot
                   signal their running configuration. Use with extra and/or
                   option to force stop with configuration options.
"""

import docopt
import yaml

from orderly_web.status import print_status
import orderly_web


def main(argv=None):
    target, args = parse_args(argv)
    target(*args)


def parse_args(argv):
    args = docopt.docopt(__doc__, argv)
    path = args["<path>"]
    if args["start"]:
        extra = args["--extra"]
        options = parse_option(args)
        pull_images = args["--pull"]
        target = orderly_web.start
        args = (path, extra, options, pull_images)
    elif args["status"]:
        target = print_status
        args = (path, )
    elif args["stop"]:
        kill = args["--kill"]
        network = args["--network"]
        volumes = args["--volumes"]
        force = args["--force"]
        extra = args["--extra"]
        options = parse_option(args)
        target = orderly_web.stop
        args = (path, kill, network, volumes, force, extra, options)
    elif args["admin"]:
        target, args = parse_admin_args(args)
    return target, args


def parse_option(args):
    return [string_to_dict(x) for x in args["--option"]]


def parse_admin_args(args):
    path = args["<path>"]
    if args["add-users"]:
        target = orderly_web.add_users
        args = (path, args["<email>"])
    elif args["add-groups"]:
        target = orderly_web.add_groups
        args = (path, args["<name>"])
    elif args["add-members"]:
        target = orderly_web.add_members
        args = (path, args["<group>"], args["<email>"])
    elif args["grant"]:
        target = orderly_web.grant
        args = (path, args["<group>"], args["<permission>"])
    return target, args


def string_to_dict(string):
    """Convert a configuration option a.b.c=x to a dictionary
{"a": {"b": "c": x}}"""
    # Won't deal with dots embedded within quotes but that's ok as
    # that should not be allowed generally.
    try:
        key, value = string.split("=")
    except ValueError:
        msg = "Invalid option '{}', expected option in form key=value".format(
            string)
        raise Exception(msg)
    value = yaml_atom_parse(value)
    for k in reversed(key.split(".")):
        value = {k: value}
    return value


def yaml_atom_parse(x):
    ret = yaml.load(x, Loader=yaml.Loader)
    if type(ret) not in [bool, int, float, str]:
        raise Exception("Invalid value '{}' - expected simple type".format(x))
    return ret



import base64
import copy
import os

import docker
import pickle
import yaml

from orderly_web.docker_helpers import docker_client, string_from_container, \
    string_into_container
import orderly_web.vault as vault

# There are two types of configuration objects and three ways that
# they turn up.  These are:

# 1. A base configuration object (`OrderlyWebConfigBase`), which
#    requires a path.  This contains only the immutable bits (which is
#    the container prefix and the container names).  This object can
#    be used to create or retrieve a full configuration.  These
#    objects are created by `read_config`
#
# 2. A full configuration object (`OrderlyWebConfig`), with all
#    options set.  This can be created by:
#
#    a. The `build()` method of the `OrderlyWebConfigBase` object, by
#       adding additional options into the base configuration.  This
#       is used when starting an OrderlyWeb constellation.  This is
#       also do-able in one step with `build_config`
#
#    b. The `fetch()` method of the `OrderlyWebConfigBase` object,
#       which retrieves a pickled configuration object from the
#       running orderly container (which stores all the options as
#       used when starting).  Also doable with `fetch_config`
#
# The idea here is that interacting with an existing set of containers
# (currently limited to status and stop, but eventually we will
# support at least upgrade too) we should not have to remember any
# additional arguments that were passed to create the container, but
# at the same time we want the startup to be configurable without
# having to edit the master configuration file.
#
# We will store a configuration into this (container, path) pair; it
# does not really matter where it is but it is ideal if it is not on a
# part of the filesystem that is persisted (i.e., not a volume)
# because it might contain secrets.
PATH_CONFIG = {"container": "orderly", "path": "/orderly-web-config"}


def read_config(path):
    return OrderlyWebConfigBase(path)


def build_config(path, extra=None, options=None):
    return read_config(path).build(extra, options)


def fetch_config(path):
    return read_config(path).fetch()


def read_yaml(filename):
    with open(filename, "r") as f:
        dat = yaml.load(f, Loader=yaml.SafeLoader)
    return dat


class OrderlyWebConfigBase:
    def __init__(self, path):
        self.path = path
        self.data = read_yaml("{}/orderly-web.yml".format(path))
        self.container_prefix = config_string(self.data, ["container_prefix"])

        if "workers" not in self.data["orderly"]:
            workers = 1
        else:
            workers = config_integer(self.data, ["orderly", "workers"])

        self.containers = {
            "redis": "{}_redis".format(self.container_prefix),
            "orderly": "{}_orderly".format(self.container_prefix),
            "web": "{}_web".format(self.container_prefix)
        }
        self.container_groups = {
            "orderly_worker": {
                "name": "{}_orderly_worker".format(self.container_prefix),
                "scale": workers
            }
        }

    def build(self, extra=None, options=None):
        data = config_data_update(self.path, self.data, extra, options)
        return OrderlyWebConfig(self.path, data)

    def fetch(self):
        try:
            with docker_client() as cl:
                name = self.containers[PATH_CONFIG["container"]]
                container = cl.containers.get(name)
        except docker.errors.NotFound:
            return None
        path = PATH_CONFIG["path"]
        txt = string_from_container(container, path)
        cfg = pickle.loads(base64.b64decode(txt))
        # We have to set the path because the relative path (or even
        # absolute path) might be different between different users of
        # the same configuration, as the docker container is a global
        # resource.
        cfg.path = self.path
        return cfg


class OrderlyWebConfig:
    def __init__(self, path, dat):
        self.path = path
        self.data = dat
        self.vault = config_vault(dat, ["vault"])
        self.network = config_string(dat, ["network"])
        self.volumes = {
            "redis": config_string(dat, ["volumes", "redis"]),
            "orderly": config_string(dat, ["volumes", "orderly"])
        }

        self.container_prefix = config_string(dat, ["container_prefix"])
        self.containers = {
            "redis": "{}_redis".format(self.container_prefix),
            "orderly": "{}_orderly".format(self.container_prefix),
            "web": "{}_web".format(self.container_prefix)
        }

        if "workers" not in dat["orderly"]:
            workers = 1
        else:
            workers = config_integer(dat, ["orderly", "workers"])

        self.container_groups = {
            "orderly_worker": {
                "name": "{}_orderly_worker".format(self.container_prefix),
                "scale": workers
            }
        }

        self.images = {
            "redis": config_image_reference(dat, ["redis", "image"]),
            "orderly": config_image_reference(dat, ["orderly", "image"]),
            "orderly_worker": config_image_reference(dat, ["orderly", "image"],
                                                     "worker_name"),
            "web": config_image_reference(dat, ["web", "image"]),
            "admin": config_image_reference(dat, ["web", "image"], "admin"),
            "migrate": config_image_reference(dat, ["web", "image"],
                                              "migrate")
        }

        self.orderly_env = config_dict(dat, ["orderly", "env"], True)

        self.web_dev_mode = config_boolean(dat, ["web", "dev_mode"], True)
        self.web_port = config_integer(dat, ["web", "port"])
        self.web_name = config_string(dat, ["web", "name"])
        self.web_email = config_string(dat, ["web", "email"])

        self.web_auth_montagu = config_boolean(
            dat, ["web", "auth", "montagu"])
        self.web_auth_fine_grained = config_boolean(
            dat, ["web", "auth", "fine_grained"])

        if not self.web_auth_montagu:
            self.web_auth_github_app = config_dict_strict(
                dat, ["web", "auth", "github_oauth"], ["id", "secret"])
            self.web_auth_github_org = config_string(
                dat, ["web", "auth", "github_org"])
            self.web_auth_github_team = config_string(
                dat, ["web", "auth", "github_team"], True)
        else:
            self.web_auth_github_app = None
            self.web_auth_github_org = None
            self.web_auth_github_team = None

        if self.web_auth_montagu:
            self.montagu_url = config_string(dat,
                                             ["web", "auth", "montagu_url"])
            self.montagu_api_url = config_string(dat,
                                                 ["web",
                                                  "auth",
                                                  "montagu_api_url"])

        self.sass_variables = config_string(dat,
                                            ["web", "sass_variables"],
                                            True)
        self.logo_path = config_string(dat, ["web", "logo"], True)
        if self.logo_path is not None:
            self.logo_path = self.get_abs_path(self.logo_path)
            self.logo_name = os.path.basename(self.logo_path)
        else:
            self.logo_name = None

        self.favicon_path = config_string(dat, ["web", "favicon"], True)
        if self.favicon_path is not None:
            self.favicon_path = self.get_abs_path(self.favicon_path)

        if self.sass_variables is not None:
            variables_abspath = self.get_abs_path(self.sass_variables)
            self.sass_variables = variables_abspath
            self.volumes["css"] = config_string(dat, ["volumes", "css"])
            css_generator = config_image_reference(dat,
                                                   ["web", "image"],
                                                   "css-generator")
            self.images["css-generator"] = css_generator

        static_documents = config_string(dat, ["volumes", "documents"], True)
        if static_documents is not None:
            self.volumes["documents"] = static_documents

        if "proxy" in dat and dat["proxy"]:
            self.proxy_enabled = config_boolean(
                dat, ["proxy", "enabled"], True)

            if self.proxy_enabled:
                self.proxy_hostname = config_string(
                    dat, ["proxy", "hostname"])
                self.proxy_port_http = config_integer(
                    dat, ["proxy", "port_http"])
                self.proxy_port_https = config_integer(
                    dat, ["proxy", "port_https"])
                ssl = config_dict(dat, ["proxy", "ssl"], True)
                self.proxy_ssl_self_signed = ssl is None
                if not self.proxy_ssl_self_signed:
                    self.proxy_ssl_certificate = config_string(
                        dat, ["proxy", "ssl", "certificate"], True)
                    self.proxy_ssl_key = config_string(
                        dat, ["proxy", "ssl", "key"], True)
                self.images["proxy"] = config_image_reference(
                    dat, ["proxy", "image"])
                self.volumes["proxy_logs"] = config_string(
                    dat, ["volumes", "proxy_logs"])
                self.containers["proxy"] = "{}_proxy".format(
                    self.container_prefix)
        else:
            self.proxy_enabled = False

        self.web_url = config_string(dat, ["web", "url"], True)
        if not self.web_url:
            if self.proxy_enabled:
                self.web_url = "https://{}:{}".format(
                    self.proxy_hostname, self.proxy_port_https)
            elif self.web_dev_mode:
                self.web_url = "http://localhost:{}".format(self.web_port)
            else:
                raise Exception("web_url must be provided")

        self.orderly_ssh = config_dict_strict(
            dat, ["orderly", "ssh"], ["public", "private"], True)

        self.orderly_initial_source = None
        self.orderly_initial_url = None
        if "initial" in dat["orderly"] and dat["orderly"]["initial"]:
            self.orderly_initial_source = config_enum(
                dat, ["orderly", "initial", "source"], ["demo", "clone"])
            if self.orderly_initial_source == "clone":
                self.orderly_initial_url = config_string(
                    dat, ["orderly", "initial", "url"])
            elif "url" in dat["orderly"]["initial"]:
                # I think an error is a bit harsh
                print("NOTE: Ignoring orderly:initial:url")

        self.slack_webhook_url = config_string(dat,
                                               ["slack", "webhook_url"],
                                               True)

    def save(self):
        orderly = self.get_container("orderly")
        txt = base64.b64encode(pickle.dumps(self)).decode("utf8")
        container = self.get_container(PATH_CONFIG["container"])
        path = PATH_CONFIG["path"]
        string_into_container(txt, container, path)

    def get_container(self, name):
        with docker_client() as cl:
            return cl.containers.get(self.containers[name])

    def resolve_secrets(self):
        vault_client = self.vault.client()
        vault.resolve_secrets(self, vault_client)
        vault.resolve_secrets(self.orderly_env, vault_client)
        vault.resolve_secrets(self.web_auth_github_app, vault_client)
        vault.resolve_secrets(self.orderly_ssh, vault_client)

    def get_abs_path(self, relative_path):
        return os.path.abspath(os.path.join(self.path, relative_path))


class DockerImageReference:
    def __init__(self, repo, name, tag):
        self.repo = repo
        self.name = name
        self.tag = tag

    def __str__(self):
        if self.repo is None:
            return "{}:{}".format(self.name, self.tag)
        else:
            return "{}/{}:{}".format(self.repo, self.name, self.tag)


def config_data_update(path, data, extra=None, options=None):
    data = copy.deepcopy(data)
    if extra:
        data_extra = read_yaml("{}/{}.yml".format(path, extra))
        config_check_additional(data_extra)
        combine(data, data_extra)
    if options:
        if type(options) == list:
            options = collapse(options)
        config_check_additional(options)
        combine(data, options)
    return data


def config_check_additional(options):
    if "container_prefix" in options:
        raise Exception("'container_prefix' may not be modified")


# Utility function for centralising control over pulling information
# out of the configuration.
def config_value(data, path, data_type, is_optional):
    if type(path) is str:
        path = [path]
    for i, p in enumerate(path):
        try:
            data = data[p]
            if data is None:
                raise KeyError()
        except KeyError as e:
            if is_optional:
                return None
            e.args = (":".join(path[:(i + 1)]),)
            raise e

    expected = {"string": str,
                "integer": int,
                "boolean": bool,
                "dict": dict}
    if type(data) is not expected[data_type]:
        raise ValueError("Expected {} for {}".format(
            data_type, ":".join(path)))
    return data


# TODO: once we have support for easily overriding parts of
# configuration, this can be made better with respect to optional
# values (e.g., if url is present other keys are required).
def config_vault(data, path):
    url = config_string(data, path + ["addr"], True)
    auth_method = config_string(data, path + ["auth", "method"], True)
    auth_args = config_dict(data, path + ["auth", "args"], True)
    return vault.vault_config(url, auth_method, auth_args)


def config_string(data, path, is_optional=False):
    return config_value(data, path, "string", is_optional)


def config_integer(data, path, is_optional=False):
    return config_value(data, path, "integer", is_optional)


def config_boolean(data, path, is_optional=False):
    return config_value(data, path, "boolean", is_optional)


def config_dict(data, path, is_optional=False):
    return config_value(data, path, "dict", is_optional)


def config_dict_strict(data, path, keys, is_optional=False):
    d = config_dict(data, path, is_optional)
    if not d:
        return None
    if set(keys) != set(d.keys()):
        raise ValueError("Expected keys {} for {}".format(
            ", ".join(keys), ":".join(path)))
    for k, v in d.items():
        if type(v) is not str:
            raise ValueError("Expected a string for {}".format(
                ":".join(path + [k])))
    return d


def config_enum(data, path, values, is_optional=False):
    value = config_string(data, path, is_optional)
    if value not in values:
        raise ValueError("Expected one of [{}] for {}".format(
            ", ".join(values), ":".join(path)))
    return value


def config_image_reference(dat, path, name="name"):
    if type(path) is str:
        path = [path]
    repo = config_string(dat, path + ["repo"], is_optional=True)
    name = config_string(dat, path + [name])
    tag = config_string(dat, path + ["tag"])
    return DockerImageReference(repo, name, tag)


def combine(base, extra):
    """Combine exactly two dictionaries recursively, modifying the first
argument in place with the contets of the second"""
    for k, v in extra.items():
        if k in base and type(base[k]) is dict and v is not None:
            combine(base[k], v)
        else:
            base[k] = v


def collapse(options):
    ret = {}
    for o in options:
        combine(ret, o)
    return ret



import os

from orderly_web.pull import pull
from orderly_web.start import start
from orderly_web.status import status
from orderly_web.stop import stop
from orderly_web.admin import add_users, add_groups, add_members, grant

__all__ = [
    pull,
    start,
    status,
    stop,
    add_users,
    add_groups,
    add_members,
    grant
]



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

requirements = [
    "docker",
    "docopt",
    "hvac",
    "pytest",
    "pyyaml",
    "vault_dev",
    "Pillow"]

setup(name="orderly_web",
      version="0.0.9",
      description="Deploy scripts for OrderlyWeb",
      long_description=long_description,
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      url="https://github.com/vimc/orderly-web-deploy",
      author="Rich FitzJohn",
      author_email="r.fitzjohn@imperial.ac.uk",
      license='MIT',
      packages=find_packages(),
      entry_points={
          'console_scripts': ['orderly-web=orderly_web.cli:main'],
      },
      include_package_data=True,
      zip_safe=False,
      # Extra:
      long_description_content_type="text/markdown",
      setup_requires=["pytest-runner"],
      tests_require=[
          "vault_dev",
          "pytest"
      ],
      install_requires=requirements)



import io
from contextlib import redirect_stdout
import pytest
import urllib
import time
import json
import ssl
import re
from unittest import mock
from urllib import request
from unittest.mock import patch, call

import requests
import vault_dev

from orderly_web.config import fetch_config, build_config
from orderly_web.docker_helpers import *
from orderly_web.errors import OrderlyWebConfigError
import orderly_web
from orderly_web.notify import Notifier


def test_status_when_not_running():
    st = orderly_web.status("config/basic")
    assert not st.is_running
    assert st.containers["orderly"]["status"] == "missing"
    assert st.containers["web"]["status"] == "missing"
    assert st.volumes == {}
    assert st.network is None


def test_status_representation_is_str():
    st = orderly_web.status("config/basic")
    f = io.StringIO()
    with redirect_stdout(f):
        print(st)
    out = f.getvalue()
    assert str(st).strip() == out.strip()
    # __repr__ is called when the object is printed
    assert st.__repr__() == str(st)


def test_start_and_stop():
    path = "config/basic"
    try:
        res = orderly_web.start(path)
        assert res
        st = orderly_web.status(path)
        assert st.containers["orderly"]["status"] == "running"
        assert st.containers["redis"]["status"] == "running"
        assert st.containers["web"]["status"] == "running"
        assert len(st.container_groups) == 1
        assert "orderly_worker" in st.container_groups
        assert st.container_groups["orderly_worker"]["scale"] == 1
        assert st.container_groups["orderly_worker"]["count"] == 1
        assert len(st.container_groups["orderly_worker"]["status"]) == 1
        assert re.match(r"orderly_web_orderly_worker_\w+",
                        st.container_groups["orderly_worker"]["status"][0]
                                           ["name"])
        assert st.container_groups["orderly_worker"]["status"][0]["status"] ==\
            "running"
        assert st.volumes["orderly"] == "orderly_web_volume"
        assert st.volumes["documents"] == "orderly_web_documents"
        assert st.volumes["redis"] == "orderly_web_redis_data"
        assert st.network == "orderly_web_network"

        f = io.StringIO()
        with redirect_stdout(f):
            res = orderly_web.start(path)
            msg = f.getvalue().strip()
        assert not res
        assert msg.endswith("please run orderly-web stop")

        cfg = fetch_config(path)
        web = cfg.get_container("web")
        ports = web.attrs["HostConfig"]["PortBindings"]
        assert list(ports.keys()) == ["8888/tcp"]
        dat = json.loads(http_get("http://localhost:8888/api/v1"))
        assert dat["status"] == "success"

        web_config = string_from_container(
            web, "/etc/orderly/web/config.properties").split("\n")

        assert "app.url=https://localhost" in web_config
        assert "auth.github_key=notarealid" in web_config
        assert "auth.github_secret=notarealsecret" in web_config
        assert "orderly.server=http://orderly_web_orderly:8321" in web_config

        # Trivial check that the proxy container works too:
        proxy = cfg.get_container("proxy")
        ports = proxy.attrs["HostConfig"]["PortBindings"]
        assert set(ports.keys()) == set(["443/tcp", "80/tcp"])
        dat = json.loads(http_get("http://localhost/api/v1"))
        assert dat["status"] == "success"
        dat = json.loads(http_get("https://localhost/api/v1"))
        assert dat["status"] == "success"

        # Orderly volume contains only the stripped down example from
        # the URL, not the whole demo:
        orderly = cfg.get_container("orderly")
        src = exec_safely(orderly, ["ls", "/orderly/src"])[1]
        src_contents = src.decode("UTF-8").strip().split("\n")
        assert set(src_contents) == set(["README.md", "example"])

        # Bring the whole lot down:
        orderly_web.stop(path, kill=True, volumes=True, network=True)
        st = orderly_web.status(path)
        assert not st.is_running
        assert st.containers["orderly"]["status"] == "missing"
        assert st.containers["redis"]["status"] == "missing"
        assert st.containers["web"]["status"] == "missing"
        assert st.container_groups["orderly_worker"]["scale"] == 1
        assert st.container_groups["orderly_worker"]["count"] == 0
        assert len(st.container_groups["orderly_worker"]["status"]) == 0
        assert st.volumes == {}
        assert st.network is None
        # really removed?
        with docker_client() as cl:
            assert not network_exists(cl, cfg.network)
            assert not volume_exists(cl, cfg.volumes["orderly"])
            assert not container_exists(cl, cfg.containers["proxy"])
    finally:
        orderly_web.stop(path, kill=True, volumes=True, network=True)


def test_start_with_custom_styles():
    path = "config/customcss"
    try:
        options = {"web": {"url": "http://localhost:8888"}}
        res = orderly_web.start(path, options=options)
        assert res
        st = orderly_web.status(path)
        assert st.containers["orderly"]["status"] == "running"
        assert st.containers["web"]["status"] == "running"
        assert st.volumes["css"] == "orderly_web_css"
        assert "documents" not in st.volumes
        assert st.network == "orderly_web_network"

        cfg = fetch_config(path)

        # check that the style volume is really mounted
        api_client = docker.client.from_env().api
        details = api_client.inspect_container(cfg.containers["web"])
        assert len(details['Mounts']) == 3
        css_volume = [v for v in details['Mounts']
                      if v['Type'] == "volume" and
                      v['Name'] == "orderly_web_css"][0]
        assert css_volume['Name'] == "orderly_web_css"
        assert css_volume['Destination'] == "/static/public/css"

        # check that the style files have been compiled with the custom vars
        web_container = cfg.get_container("web")
        style = string_from_container(web_container,
                                      "/static/public/css/style.css")
        assert "/* Example custom config */" in style

        # check that js files are there also
        res = requests.get("http://localhost:8888/js/index.bundle.js")
        assert res.status_code == 200

        # check that the custom logo is mounted and appears on the page
        logo_mount = [v for v in details['Mounts']
                      if v['Type'] == "bind"][0]
        expected_destination = "/static/public/img/logo/my-test-logo.png"
        assert logo_mount['Destination'] == expected_destination
        res = requests.get("http://localhost:8888")
        assert """<img src="http://localhost:8888/img/logo/my-test-logo.png"""\
               in res.text
        res = requests.get("http://localhost:8888/img/logo/my-test-logo.png")
        assert res.status_code == 200
    finally:
        orderly_web.stop(path, kill=True, volumes=True, network=True)


def test_stop_broken_orderly_web():
    path = "config/breaking"
    try:
        start_failed = False
        try:
            orderly_web.start(path)
        except docker.errors.APIError:
            start_failed = True

        assert start_failed

        stop_failed = False
        try:
            orderly_web.stop("config/breaking")
        except OrderlyWebConfigError:
            stop_failed = True

        assert stop_failed

        with docker_client() as cl:
            assert container_exists(cl, "orderly_web_orderly")
    finally:
        orderly_web.stop(path, force=True, network=True, volumes=True)
        with docker_client() as cl:
            assert not container_exists(cl, "orderly_web_orderly")


def test_stop_broken_orderly_web_with_option():
    path = "config/breaking"
    options = [{"network": "ow_broken_test"}]
    try:
        start_failed = False
        try:
            orderly_web.start(path, options=options)
        except docker.errors.APIError:
            start_failed = True

        assert start_failed

        with docker_client() as cl:
            assert container_exists(cl, "orderly_web_orderly")
            assert network_exists(cl, "ow_broken_test")
    finally:
        orderly_web.stop(path, force=True, network=True, volumes=True,
                         options=options)
        with docker_client() as cl:
            assert not container_exists(cl, "orderly_web_orderly")
            assert not network_exists(cl, "ow_broken_test")


def test_stop_broken_orderly_web_with_extra():
    path = "config/breaking"
    extra = "extra"  # defines network as "ow_broken_extra_test"
    try:
        start_failed = False
        try:
            orderly_web.start(path, extra=extra)
        except docker.errors.APIError:
            start_failed = True

        assert start_failed

        with docker_client() as cl:
            assert container_exists(cl, "orderly_web_orderly")
            assert network_exists(cl, "ow_broken_extra_test")
    finally:
        orderly_web.stop(path, force=True, network=True, volumes=True,
                         extra=extra)
        with docker_client() as cl:
            assert not container_exists(cl, "orderly_web_orderly")
            assert not network_exists(cl, "ow_broken_extra_test")


def test_status_from_broken_orderly_web():
    path = "config/breaking"
    try:
        start_failed = False
        try:
            orderly_web.start(path)
        except docker.errors.APIError:
            start_failed = True

        assert start_failed

        status = orderly_web.status(path)
        assert str(status) == "Cannot read status from orderly-web because " \
            "it has not started successfully or is in an error state."
    finally:
        orderly_web.stop(path, force=True, network=True, volumes=True)


def test_start_with_montagu_config():
    path = "config/montagu"
    try:
        res = orderly_web.start(path)
        assert res
        st = orderly_web.status(path)
        assert st.containers["orderly"]["status"] == "running"
        assert st.containers["web"]["status"] == "running"
        assert st.network == "orderly_web_network"

        cfg = fetch_config(path)
        web = cfg.get_container("web")
        web_config = string_from_container(
            web, "/etc/orderly/web/config.properties").split("\n")

        assert "montagu.url=http://montagu" in web_config
        assert "montagu.api_url=http://montagu/api" in web_config
        assert 'auth.github_org=' in web_config
        assert 'auth.github_team=' in web_config

    finally:
        orderly_web.stop(path, kill=True, volumes=True, network=True)


def test_admin_cli():
    path = "config/basic"
    try:
        orderly_web.start(path)
        result = orderly_web.add_users(path, ["test.user@gmail.com"])
        expected = "Saved user with email 'test.user@gmail.com' to the " \
                   "database"
        assert expected in result
        result = orderly_web.add_groups(path, ["funders"])
        assert "Saved user group 'funders' to the database" in result
        result = orderly_web.add_members(path, "funders",
                                         ["test.user@gmail.com"])
        expected = "Added user with email 'test.user@gmail.com' to user " \
                   "group 'funders'"
        assert expected in result
        result = orderly_web.grant(path, "funders", ["*/reports.read"])
        expected = "Gave user group 'funders' the permission '*/reports.read'"
        assert expected in result
        result = orderly_web.grant(path, "funders", ["*/nonsense"])
        assert "Unknown permission : 'nonsense'" in result
    finally:
        orderly_web.stop(path, kill=True, volumes=True, network=True)


def test_no_devmode_no_ports():
    path = "config/noproxy"
    options = {"web": {"dev_mode": False,
                       "url": "http://localhost"}}

    try:
        orderly_web.start(path, options=options)
        cfg = fetch_config(path)
        assert not cfg.web_dev_mode
        web = cfg.get_container("web")
        assert web.attrs["HostConfig"]["PortBindings"] is None
    finally:
        orderly_web.stop(path, kill=True, volumes=True, network=True)


def test_can_pull_on_deploy():
    path = "config/noproxy"
    migrate_image = orderly_web.config.build_config(path).images["migrate"]
    with docker_client() as cl:
        try:
            cl.images.remove(str(migrate_image), noprune=True)
        except docker.errors.ImageNotFound:
            pass
        res = orderly_web.start(path, pull_images=True)
        img = cl.images.get(str(migrate_image))
        assert str(migrate_image) in img.tags
        orderly_web.stop(path, kill=True, volumes=True, network=True)


def test_vault_ssl():
    with vault_dev.server() as s:
        cl = s.client()
        # Copy the certificates into the vault where we will later on
        # pull from from.
        cert = read_file("proxy/ssl/certificate.pem")
        key = read_file("proxy/ssl/key.pem")
        cl.write("secret/ssl/certificate", value=cert)
        cl.write("secret/ssl/key", value=key)
        cl.write("secret/db/password", value="s3cret")
        cl.write("secret/github/id", value="ghid")
        cl.write("secret/github/secret", value="ghs3cret")
        cl.write("secret/ssh", public="public-key-data",
                 private="private-key-data")
        cl.write("secret/slack/webhook", value="http://webhook")

        path = "config/complete"

        vault_addr = "http://localhost:{}".format(s.port)
        vault_auth = {"args": {"token": s.token}}
        options = {"vault": {"addr": vault_addr, "auth": vault_auth}}

        res = orderly_web.start(path, options=options)
        dat = json.loads(http_get("https://localhost/api/v1"))
        assert dat["status"] == "success"

        cfg = fetch_config(path)
        container = cfg.get_container("orderly")
        res = string_from_container(container, "/root/.Renviron")
        assert "ORDERLY_DB_PASS=s3cret" in res

        private = string_from_container(container, "/root/.ssh/id_rsa")
        assert private == "private-key-data"
        public = string_from_container(container, "/root/.ssh/id_rsa.pub")
        assert public == "public-key-data"

        known_hosts = string_from_container(container,
                                            "/root/.ssh/known_hosts")
        assert "github.com" in known_hosts

        web_container = cfg.get_container("web")
        web_config = string_from_container(
            web_container,
            "/etc/orderly/web/config.properties").split("\n")

        assert "auth.github_key=ghid" in web_config
        assert "auth.github_secret=ghs3cret" in web_config

        orderly_web.stop(path, kill=True, volumes=True, network=True)


def test_without_github_app_for_montagu():
    path = "config/basic"
    options = {"web": {"auth": {"montagu": True,
                                "montagu_url": "montagu",
                                "montagu_api_url": "montagu/api",
                                "github_key": None,
                                "github_secret": None}}}
    res = orderly_web.start(path, options=options)
    assert res
    st = orderly_web.status(path)
    assert st.containers["orderly"]["status"] == "running"
    assert st.containers["web"]["status"] == "running"
    assert st.network == "orderly_web_network"

    orderly_web.stop(path, kill=True, volumes=True, network=True)


# To run this test you will need a token for the vimc robot user -
# this can be found in the montagu vault as
# /secret/vimc-robot/vault-token
# This environment variable is configured on travis
def test_vault_github_login_with_prompt():
    if "VAULT_AUTH_GITHUB_TOKEN" in os.environ:
        del os.environ["VAULT_AUTH_GITHUB_TOKEN"]
    with mock.patch('builtins.input',
                    return_value=os.environ["VAULT_TEST_GITHUB_PAT"]):
        with vault_dev.server() as s:
            cl = s.client()
            enable_github_login(cl)
            cl.write("secret/db/password", value="s3cret")

            path = "config/vault"
            vault_addr = "http://localhost:{}".format(s.port)
            options = {"vault": {"addr": vault_addr}}

            orderly_web.start(path, options=options)

            cfg = fetch_config(path)
            container = cfg.get_container("orderly")
            res = string_from_container(container,
                                        "/root/.Renviron")
            assert "ORDERLY_DB_PASS=s3cret" in res

            orderly_web.stop(path, kill=True, volumes=True, network=True)


# To run this test you will need a token for the vimc robot user -
# this can be found in the montagu vault as
# /secret/vimc-robot/vault-token
# This environment variable is configured on travis
def test_vault_github_login_from_env():
    os.environ["VAULT_AUTH_GITHUB_TOKEN"] = os.environ["VAULT_TEST_GITHUB_PAT"]
    with vault_dev.server() as s:
        cl = s.client()
        enable_github_login(cl)
        cl.write("secret/db/password", value="s3cret")

        path = "config/vault"
        vault_addr = "http://localhost:{}".format(s.port)
        options = {"vault": {"addr": vault_addr}}

        orderly_web.start(path, options=options)

        cfg = fetch_config(path)
        container = cfg.get_container("orderly")
        res = string_from_container(container,
                                    "/root/.Renviron")
        assert "ORDERLY_DB_PASS=s3cret" in res

        orderly_web.stop(path, kill=True, volumes=True,
                         network=True)


# To run this test you will need a token for the vimc robot user -
# this can be found in the montagu vault as
# /secret/vimc-robot/vault-token
# This environment variable is configured on travis
def test_vault_github_login_with_mount_path():
    os.environ["VAULT_AUTH_GITHUB_TOKEN"] = os.environ["VAULT_TEST_GITHUB_PAT"]
    with vault_dev.server() as s:
        cl = s.client()
        enable_github_login(cl, path="github-custom")
        cl.write("secret/db/password", value="s3cret")

        path = "config/vault"
        vault_addr = "http://localhost:{}".format(s.port)
        options = {"vault":
                   {"addr": vault_addr,
                    "auth":
                    {"method": "github",
                     "args": {"mount_point": "github-custom"}}}}

        orderly_web.start(path, options=options)

        cfg = fetch_config(path)
        container = cfg.get_container("orderly")
        res = string_from_container(container,
                                    "/root/.Renviron")
        assert "ORDERLY_DB_PASS=s3cret" in res

        orderly_web.stop(path, kill=True, volumes=True,
                         network=True)


def test_error_if_orderly_not_initialised():
    path = "config/basic"
    options = {"orderly": {"initial": None}}
    cfg = build_config(path, options=options)
    # ensure this test behaves sensibly if state is a bit messy:
    with docker_client() as cl:
        remove_volume(cl, cfg.volumes["orderly"])
    try:
        with pytest.raises(Exception,
                           match="Orderly volume not initialised"):
            res = orderly_web.start(path, options=options)
    finally:
        orderly_web.stop(path, force=True, network=True, volumes=True,
                         kill=True)


def test_can_start_with_prepared_volume():
    path = "config/basic"
    options = {"orderly": {"initial": None}}
    cfg = build_config(path, options=options)

    # ensure this test behaves sensibly if state is a bit messy:
    with docker_client() as cl:
        remove_volume(cl, cfg.volumes["orderly"])
        image = str(cfg.images["orderly"])
        mounts = [docker.types.Mount("/orderly", cfg.volumes["orderly"])]
        args = ["Rscript", "-e", "orderly:::create_orderly_demo('/orderly')"]
        cl.containers.run(
            image, entrypoint=args, mounts=mounts, auto_remove=True)

    try:
        f = io.StringIO()
        with redirect_stdout(f):
            res = orderly_web.start(path, options=options)
        assert res
        out = f.getvalue()
        expected = 'orderly volume already contains data - not initialising'
        assert expected in out.splitlines()
    finally:
        orderly_web.stop(path, kill=True, volumes=True, network=True)


def test_notifies_slack_on_success():
    with patch.object(Notifier, 'post',
                      return_value=None) as mock_notify:
        path = "config/basic"
        try:
            orderly_web.start(path)
        finally:
            orderly_web.stop(path, kill=True, volumes=True, network=True)
    calls = [call("*Starting* deploy to https://localhost"),
             call("*Completed* deploy to https://localhost :shipit:")]
    mock_notify.assert_has_calls(calls)


def test_notifies_slack_on_fail():
    with patch.object(Notifier, 'post',
                      return_value=None) as mock_notify:
        path = "config/breaking"
        try:
            orderly_web.start(path)
        except docker.errors.APIError:
            start_failed = True
        finally:
            orderly_web.stop(path, force=True, network=True, volumes=True)
    calls = [call("*Starting* deploy to https://localhost"),
             call("*Failed* deploy to https://localhost :bomb:")]
    mock_notify.assert_has_calls(calls)


def test_start_and_stop_multiple_workers():
    options = {"orderly": {"workers": 2}}
    path = "config/basic"
    try:
        res = orderly_web.start(path, options=options)
        assert res
        st = orderly_web.status(path)
        assert st.containers["orderly"]["status"] == "running"
        assert st.containers["redis"]["status"] == "running"
        assert st.containers["web"]["status"] == "running"
        assert len(st.container_groups) == 1
        assert "orderly_worker" in st.container_groups
        assert st.container_groups["orderly_worker"]["count"] == 2
        assert len(st.container_groups["orderly_worker"]["status"]) == 2
        assert re.match(r"orderly_web_orderly_worker_\w+",
                        st.container_groups["orderly_worker"]["status"][0]
                                           ["name"])
        assert st.container_groups["orderly_worker"]["status"][0]["status"] ==\
            "running"
        assert re.match(r"orderly_web_orderly_worker_\w+",
                        st.container_groups["orderly_worker"]["status"][1]
                                           ["name"])
        assert st.container_groups["orderly_worker"]["status"][1]["status"] ==\
            "running"

        # Bring the whole lot down:
        orderly_web.stop(path, kill=True, volumes=True, network=True)
        st = orderly_web.status(path)
        assert not st.is_running
        assert st.containers["orderly"]["status"] == "missing"
        assert st.containers["redis"]["status"] == "missing"
        assert st.containers["web"]["status"] == "missing"
        assert st.container_groups["orderly_worker"]["count"] == 0
        assert len(st.container_groups["orderly_worker"]["status"]) == 0
    finally:
        orderly_web.stop(path, kill=True, volumes=True, network=True)


def enable_github_login(cl, path="github"):
    cl.sys.enable_auth_method(method_type="github", path=path)
    policy = """
           path "secret/*" {
             capabilities = ["read", "list"]
           }
           """

    cl.sys.create_or_update_policy(
        name='secret-reader',
        policy=policy,
    )

    cl.auth.github.map_team(
        team_name="robots",
        policies=["secret-reader"],
        mount_point=path
    )

    cl.auth.github.configure(organization="vimc", mount_point=path)


# Because we wait for a go signal to come up, we might not be able to
# make the request right away:
def http_get(url, retries=5, poll=0.5):
    ctx = ssl.create_default_context()
    ctx.check_hostname = False
    ctx.verify_mode = ssl.CERT_NONE
    for i in range(retries):
        try:
            r = urllib.request.urlopen(url, context=ctx)
            return r.read().decode("UTF-8")
        except (urllib.error.URLError, ConnectionResetError) as e:
            print("sleeping...")
            time.sleep(poll)
            error = e
    raise error


def read_file(path):
    with open(path, "r") as f:
        return f.read()



import docker

from orderly_web.config import fetch_config
from orderly_web.docker_helpers import docker_client, return_logs_and_remove


def add_users(path, emails):
    args = ["add-users"] + emails
    return run(path, args)


def add_groups(path, names):
    args = ["add-groups"] + names
    return run(path, args)


def add_members(path, name, emails):
    args = ["add-members", name] + emails
    return run(path, args)


def grant(path, name, permissions):
    args = ["grant", name] + permissions
    return run(path, args)


def run(path, args):
    cfg = fetch_config(path)
    image = str(cfg.images["admin"])
    with docker_client() as cl:
        mounts = [docker.types.Mount("/orderly", cfg.volumes["orderly"])]
        result = return_logs_and_remove(cl, image, args, mounts)
        print(result)
        return result
